/*
  Copyright 2003 John Knottenbelt
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/

#include "config.h"
#include "dvbd.h"
#include "utils.h"
#include "stringutil.h"
#include "select.h"
#include "clientconnection.h"
#include "unixclientsocket.h"

#include <string>
#include <iostream>
#include <time.h>
#include <getopt.h>
#include <cerrno>
#include <fstream>
#include <fcntl.h>

void usage(const char *progname);
time_t convertTime(const char *theTime);
int openOutputFile(std::string &outputfile);
void closeOutputFile(int fd);

int main(int argc, char *argv[])
{
  // improve iostream performance
  std::ios::sync_with_stdio(false);

  std::string outputFile = "-";
  std::string socketFile = DEFAULT_SOCKET_FILE;
  bool durationSpecified = false;
  unsigned duration = 0;
  int priority = 10;
  bool convertToPS = false, convertToES = false;

  while (true) {
    static struct option long_options[] = {
      {"duration", 1, 0, 'd'},
      {"priority", 1, 0, 'p'},
      {"output", 1, 0, 'o'},
      {"socket", 1, 0, 's'},
      {"ps", 0, 0, 'P'},
      {"es", 0, 0, 'E'},
      {0, 0, 0, 0}
    };
 
    int option_index = 0;
    int c = getopt_long(argc, argv, "EPd:p:o:s:", long_options, &option_index);
    if (c == -1)
      break;
    switch (c) {
    case 'd':
      durationSpecified = true;
      duration = toInt(optarg);
      break;
    case 'p':
      priority = toInt(optarg);
      break;
    case 'o':
      outputFile = optarg;
      break;
    case 's':
      socketFile = optarg;
      break;
    case 'E':
      convertToES = true;
      break;
    case 'P':
      convertToPS = true;
      break;
    default:
      usage(argv[0]);
      return 1;
    }
  }

  if (argc - optind != 2) {
    if (argc - optind != 0)
      std::cerr << "Incorrect number of arguments passed (expected 2, got " 
		<< argc - optind << ")\n\n";

    usage(argv[0]);
    return 1;
  }

  // Parse the command line arguments
  std::string type(argv[optind++]);
  std::string channel(argv[optind++]);

  ClientConnection client(socketFile);
  if (!client) {
    std::cerr << "Fatal error: failed to connect to server on socket " << socketFile << std::endl;
    return 1;
  }

  // Open the output file
  int outputFd = openOutputFile(outputFile);
  if (outputFd < 0) {
    std::cerr << "Fatal error: failed to open " << outputFile << " for output"
	      << std::endl;
    return 1;
  }

  // Tune the channel
  if (!client.tune(type, channel, convertToPS, priority)) {
    std::cerr << "Failed to tune channel. Response from server was:\n"
	      << client.getError() << "\n";
    return 1;
  }

  UnixClientSocket &control = client.control();
  UnixClientSocket &data = client.data();

  control.setBlocking(false);
  data.setBlocking(false);

  time_t start = time(NULL);
  Select s;
  do {
    s.clearFDs();
    control.addSelectFDs(s);
    data.addSelectFDs(s);

    if (durationSpecified) {
      unsigned elapsed = time(NULL) - start;
      if (elapsed >= duration*60) {
	std::cerr << "Duration elapsed" << std::endl;
	break;
      }
      s.addAlarm(duration * 60 - elapsed, 0);
    }

    s.wait();

    if (data.isReady(s)) {
      if (!data.processReady(s)) {
	std::cerr << "Server closed data connection" << std::endl;
	break;
      }

      char buffer[1024];
      int bytesRead = data.read(buffer, 1024);

      if (bytesRead > 0) 
	write(outputFd, buffer, bytesRead);
    }
    else if (control.isReady(s)) {
      if (!control.processReady(s)) {
	std::cerr << "Server closed control connection" << std::endl;
	break;
      }

      std::string line;
      if (control.readLine(line)) 
	std::cerr << line;
    }
  } while (true);

  closeOutputFile(outputFd);
  return 0;
}

int openOutputFile(std::string &outputFile)
{
  if (outputFile == "-") 
    return 1;

  int openMode = O_WRONLY|O_CREAT|O_TRUNC|O_LARGEFILE;

  // Don't bother caching if possible
#ifdef O_STREAMING
  openMode |= O_STREAMING;
#endif
  
  return open(outputFile.c_str(), openMode, 0644);
}

void closeOutputFile(int fd)
{
  if (fd >= 0)
    close(fd);
}

void usage(const char *progname) {
  std::cerr << "Usage: " << progname << " [OPTION] TYPE CHANNEL\n"
	    << "Receive DVB broadcast for channel CHANNEL, using DVB-card\n"
	    << "\n"
	    << "TYPE can be dvb-t, dvb-s or dvb-c.\n"
	    << "\n"
	    << "  -d, --duration         Time to stream for in minutes.\n"
	    << "  -p, --priority         Priority of stream (default is 10\n"
	    << "                         higher priorities may dislodge lower\n"
	    << "                         ones).\n"
	    << "  -o, --output           Record to output file\n"
	    << "                         (default is standard output).\n"
	    << "  -s, --socket           Specify socket to connect to\n"
	    << "                         (default is " << DEFAULT_SOCKET_FILE << ")\n"
	    << "  -P, --ps               Convert stream to program stream\n"
	    << "  -E, --es               Convert stream to elementary stream (audio only)\n";
}
